#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''Quantum Resource Analysis Tools.'''
import numpy as np
from dataclasses import dataclass
from .qec import SurfaceCode
from .qdevices import QPU
from .qdecoding import (_DecoderAlgorithm, _DecoderHardware)
from .qproblems import LogicalCircuit
from .util import data_rate_heuristic


@dataclass
class QuantumSystem:
    '''Full quantum system to analyze consisting of:

        * quantum circuit
        * qpu
        * quantum error correction code
        * decoding algorithm
        * decoding hardware
    '''
    circuit: LogicalCircuit
    qpu: QPU
    code: SurfaceCode
    dec_algo: _DecoderAlgorithm
    dec_hw: _DecoderHardware


def hscx_cycle_time(d: int, qsystem: QuantumSystem):
    '''Compute the (logical) cycle time of a HSCX sequence in s.

    Args
    ----
        d (int):
            code distance
        qsystem (QuantumSystem):
            full quantum system

    Returns
    -------
        time in seconds for HSCX-sequence
    '''
    return 10 * d * qsystem.qpu.code_cycle_duration


def t_cycle_time(d: int, M: int, qsystem: QuantumSystem):
    '''

    Args
    ----
        d (int):
            code distance
        M (int):
            stopping time in us
        qsystem (QuantumSystem):
            full quantum system

    Returns
    -------
        delta_t (float):
            time in seconds for T-sequence
        t_msd (float):
            time for MSD
        t_route (float):
            time for route + CX + meas
        t_dec (float):
            time for decoding
        t_latency (float):
            time for latency
    '''
    t_cyc = qsystem.qpu.code_cycle_duration
    ep = qsystem.qpu.physical_error
    t_msd = qsystem.code.cycles_per_magic_state * d * t_cyc
    t_route = 7 * d * t_cyc
    t_dec = 2 * (qsystem.dec_algo.max_runtime(ep, d, M) / qsystem.dec_hw.speedup) # depth-2 parallel window # noqa
    t_latency = 2 * qsystem.dec_hw.latency
    delta_t = t_msd + t_route + t_dec + t_latency
    return delta_t, t_msd, t_route, t_dec, t_latency


def logical_clock_speed(d: int, M: int, qsystem: QuantumSystem):
    '''
    Compute logical clock speed.

    Args
    ----
        d (int):
            code distance
        M (int):
            stopping time in us
        qsystem (QuantumSystem):
            full quantum system

    Returns
    -------
        logical clock speed in Hz
    '''
    t_tot = hscx_cycle_time(d, qsystem) + t_cycle_time(d, M, qsystem)[0]
    return 1./t_tot


def optimize_code_distance(
    error_budget: float, M: int, qsystem: QuantumSystem,
    d_range=np.arange(3, 101, 2)
):
    '''Compute the required code distance to solve the problem with the given
    error budget. This implementation uses a straightforward search procedure.

    Args
    ----
        error_budget (float):
            total error budget for this quantum computation (value between
            (0, 1))
        M (int):
            stopping time in us
        qsystem (QuantumSystem):
            full quantum system specification
        d_range (Iterable):
            array with code distances d to search over (assumed to be array
            of increasing odd integers.)

    Returns
    ------
        the minimally required code distance d (None if no solution),
        the error at this code distance (None if no solution),
        all errors of the range of code distances d
    '''
    errors = np.zeros(d_range.shape)
    for i, d in enumerate(d_range):
        errors[i] = compute_logical_error(
            d=d,
            M=M,
            qsystem=qsystem
        )
    opt_i = next(
        (i for i in range(len(errors)) if errors[i] < error_budget), None
    )
    if opt_i is not None:
        d_opt = d_range[opt_i]
        err_opt = errors[opt_i]
        return d_opt, err_opt, errors
    return None, None, errors


def optimize_interruption_time_and_code_distance(
        error_budget: float, qsystem: QuantumSystem,
        d_range=np.arange(3, 33, 2)
):
    '''Compute the optimal code distance and interruption time.

    Args
    ----
        error_budget (float):
            error to be tolerated
        qsystem (QuantumSystem):
            full quantum system specification
        d_range
    '''
    M, it, found_ub = 1, 0, False
    max_it = 20
    while not found_ub and it < max_it:
        d, _, _ = optimize_code_distance(error_budget, M, qsystem, d_range)
        if d is None:
            it += 1
            M *= 2
        else:
            found_ub = True
    if d is None:
        raise RuntimeError(
            'No valid solution found.'
        )
    # refine estimate
    ub, lb = M, M/2
    max_it = it + 10
    it = 0
    found = False
    while not found and it < max_it:
        if ub - lb < 2:
            found = True
        else:
            mid = int((ub + lb) / 2)
            d, _, _ = optimize_code_distance(
                error_budget, mid, qsystem, d_range
            )
            it += 1
            if d is None:
                # mid -> ub
                lb = mid
            else:
                # mid -> lb
                ub = mid
    d, err, errors = optimize_code_distance(error_budget, ub, qsystem, d_range)
    return ub, d, err, errors


def compute_logical_error(
        d: int, M: int, qsystem: QuantumSystem
):
    '''Compute the total logical error rate of a full quantum computation.

    Args
    ----
        d (int):
            code distance
        M (int):
            stopping time in us
        qsystem (QuantumSystem):
            full quantum system

    Returns
    ------
        logical error rate
    '''
    ep = qsystem.qpu.physical_error
    num_patch = qsystem.code.compute_total_number_of_tiles(
        qsystem.circuit.qubits
    )
    edec = qsystem.dec_algo.decoding_error_rate(ep, d, M)
    delta = hscx_cycle_time(d, qsystem) + t_cycle_time(d, M, qsystem)[0]
    return edec * num_patch * qsystem.circuit.tgates * delta


def compute_nt_range(error_budget, d, M, qsystem: QuantumSystem):
    '''Compute the T-gate range of the decoder.

    Args
    ----
        error_budget (float):
            total error budget for this quantum computation (value between
            (0, 1))
        d (int):
            code distance
        M (int):
            stopping time in us
        qsystem (QuantumSystem):
            full quantum system

    Returns
    -------
        nt (Int):
            the number of T-gates that can be reliably run within the error
            budget
    '''
    nt, it, found_ub = 1, 0, False
    max_it = 40
    circuit = LogicalCircuit(name='tmp', qubits=1, tgates=nt)
    while not found_ub and it < max_it:
        # we compute the error for 1 patch / logical qubit
        edec = qsystem.dec_algo.decoding_error_rate(
            qsystem.qpu.physical_error, d, M
        )
        delta = hscx_cycle_time(d, qsystem) + t_cycle_time(d, M, qsystem)[0]
        err = edec * circuit.tgates * delta
        it += 1
        if err > error_budget:
            found_ub = True
        else:
            nt *= 2
            circuit.tgates = nt

    ub, lb = nt, nt/2
    max_it = it + 1
    it = 0
    found = False
    while not found and it < max_it:
        if ub - lb < 2:
            found = True
        else:
            mid = int((ub + lb) / 2)
            circuit.tgates = mid
            # we compute the error for 1 patch / logical qubit
            edec = qsystem.dec_algo.decoding_error_rate(
                qsystem.qpu.physical_error, d, M
            )
            delta = hscx_cycle_time(d, qsystem) + \
                t_cycle_time(d, M, qsystem)[0]
            err = edec * circuit.tgates * delta
            it += 1
            if err > error_budget:
                # mid -> ub
                ub = mid
            else:
                # mid -> lb
                lb = mid
    return ub


def compute_runtime(
        d: int, M: int, qsystem: QuantumSystem
):
    '''
    Compute the runtime of the quantum computation in seconds.

    Args
    ----
        d (int):
            code distance
        M (int):
            stopping time in us
        qsystem (QuantumSystem):
            full quantum system

    Return
    ------
        runtime in seconds
    '''
    delta = hscx_cycle_time(d, qsystem) + t_cycle_time(d, M, qsystem)[0]
    return qsystem.circuit.tgates * delta


def compute_nwindows(
        d: int, M: int, qsystem: QuantumSystem
):
    '''Compute the number of parallel windows required for real-time parallel
     window decoding on this quantum computation.

    Args
    ----
        d (int):
            code distance
        M (int):
            stopping time in us
        qsystem (QuantumSystem):
            full quantum system
    '''
    delta = hscx_cycle_time(d, qsystem) + t_cycle_time(d, M, qsystem)[0]
    nwindows_one_patch = np.ceil(
        np.ceil(delta/qsystem.qpu.code_cycle_duration) / (4*d)
    )
    num_patch = qsystem.code.compute_total_number_of_tiles(
        qsystem.circuit.qubits
    )
    return nwindows_one_patch * num_patch


def compute_data_rate(
        d: int, qsystem: QuantumSystem
):
    '''Compute the raw and compressed data generation rates of this quantum
    computation in bits/seconds.


    Args
    ----
        d (int):
            code distance
        qsystem (QuantumSystem):
            full quantum system

    Returns
    -------
        raw_rate (float):
            raw data generation rate in bps
        compr_rate (float)
            compressed data generation rate in bps
    '''
    # circuit: LogicalCircuit, qpu: QPU, code: SurfaceCode
    compr_rate = data_rate_heuristic(
        qsystem.qpu.physical_error, d, qsystem.qpu.code_cycle_duration
    )
    raw_rate = d**2 / qsystem.qpu.code_cycle_duration
    num_patch = qsystem.code.compute_total_number_of_tiles(
        qsystem.circuit.qubits
    )
    return raw_rate * num_patch, compr_rate * num_patch
