#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''Models of quantum syndrome decoders.'''
from typing import Callable
import scipy as sp
import numpy as np
from .util import decoding_error_rate_heuristic
from dataclasses import dataclass


class _DecoderAlgorithm:
    '''Stochastic model for the runtime of a decoder algorithm.'''
    def __init__(
            self, name: str, gen_runtime_params: Callable,
            failure_amplification: float
    ) -> None:
        '''

        Args
        ----
            name (str): Name for decoder model
            gen_runtime_params (Callable): generate the parameters for the
            runtime sampler in this model
            failure_amplification (float): amplification factor on the
            heuristic Fowler rate.
        '''
        self.name = name
        self._gen_runtime_params = gen_runtime_params
        self.failure_amplification = failure_amplification

    def _runtime_sampler(self, ep: float, d: int):
        '''Generate the runtime sampler for the decoder.

        Args
        ----
            ep (Float): physical error rate
            d (Int): code distance

        Returns
        -------
            stochastic (binomial) random variable that allows to sample from
            the runtime. (in us)
        '''
        n, p = self._gen_runtime_params(ep, d)
        return sp.stats.binom(n=int(np.ceil(n)), p=p)

    def max_runtime(self, ep: float, d: int, M: int = None):
        '''Compute the maximum runtime of this decoder.

        Args
        ---
            n (int): number of samples
            ep (Float): physical error rate
            d (Int): code distance
            M (int): interruption time (in us)
        Returns
        ------
            the maximum runtime in seconds
        '''
        n, _ = self._gen_runtime_params(ep, d)
        if M:
            return 1e-6 * min(n, M)
        return 1e-6 * n

    def generate_runtimes(self, n: int, ep: float, d: int, M: int = None):
        '''Generate n runtime samples based on the model.

        Args
        ---
            n (int): number of samples
            ep (Float): physical error rate
            d (Int): code distance
            M (int): interruption time (in us)

        Returns
        -------
            array with n runtime samples (in seconds)
        '''
        samples = self._runtime_sampler(ep, d).rvs(n)
        if M:
            samples[samples > M] = np.inf
        return 1e-6 * samples

    def pmf(self, k: int, ep: float, d: int, M: int = None):
        '''Probability mass function.

        Args
        ----
            k (int): number of 1 us intervals
            ep (Float): physical error rate
            d (Int): code distance
            M (int): interruption time (in us)
        Returns
        -------
            probability that decoder takes k intervals to complete
        '''
        _pmf = self._runtime_sampler(ep, d).pmf(k)
        if M:
            return _pmf / self._runtime_sampler(ep, d).cdf(M)
        return _pmf

    def cdf(self, k: int, ep: float, d: int, M: int = None):
        '''Cumulative distribution function.

        Args
        ----
            k (int): number of 1 us intervals
            ep (Float): physical error rate
            d (Int): code distance
            M (int): interruption time (in us)
        Returns
        -------
            probability that decoder takes less than or equal to M us
        '''
        _cdf = self._runtime_sampler(ep, d).cdf(k)
        if M:
            return min(_cdf / self._runtime_sampler(ep, d).cdf(M), 1.0)
        return _cdf

    def decoding_error_rate(self, ep: float, d: int, M: int = None):
        '''Scaled Fowler & Gidney heuristic for this decoder.

        Args
        ----
            ep (Float):           physical error rate
            d (Int):              code distance
            M (int): interruption time (in us)
        Returns
        -------
            el (Float):           logical error rate of decoder
        '''
        el = self.failure_amplification * decoding_error_rate_heuristic(ep, d)

        if M:
            return el * self._runtime_sampler(ep, d).cdf(M) + \
                ep * self._runtime_sampler(ep, d).sf(M)
        return el


def _gen_linear_runtime_params(ep: float, d: int):
    '''Generate parameters for linearly scaling decoder.

    Args
    ----
        ep (Float): physical error rate
        d (Int): code distance

    Returns
    -------
        n, p parameters for linearly scaling binomial distribution
    '''
    n = 0.25*d**3
    p = ep
    return n, p


_linear = _DecoderAlgorithm(
    name='linear',
    gen_runtime_params=_gen_linear_runtime_params,
    failure_amplification=4./3.
)


def _gen_quadratic_runtime_params(ep: float, d: int):
    '''Generate parameters for quadratically scaling decoder.

    Args
    ----
        ep (Float): physical error rate
        d (Int): code distance

    Returns
    -------
        n, p parameters for quadratically scaling binomial distribution (in us)
    '''
    n = d**6
    p = ep/d**3
    return n, p


_quadratic = _DecoderAlgorithm(
    name='quadratic',
    gen_runtime_params=_gen_quadratic_runtime_params,
    failure_amplification=1.
)


def _gen_sparse_blossom_params(ep: float, d: int):
    '''Generate parameters for O(N^1.2) scaling of sparse blossom.

    Args
    ----
        ep (Float): physical error rate
        d (Int): code distance

    Returns
    -------
        n, p parameters for  binomial distribution matching experimental
        scaling (in us)
    '''
    n = 6.1*d**4.3
    p = ep
    return n, p


_sparse_blossom = _DecoderAlgorithm(
    name="sparse blossom",
    gen_runtime_params=_gen_sparse_blossom_params,
    failure_amplification=1.
)


@dataclass
class _Algorithms:
    linear: _DecoderAlgorithm = _linear
    quadratic: _DecoderAlgorithm = _quadratic
    sparseblossom: _DecoderAlgorithm = _sparse_blossom


algorithms = _Algorithms()


class _DecoderHardware:
    '''Model for the classical hardware running the decoder.'''
    def __init__(
            self, name: str, speedup: float = 1, parallelism: int = 1,
            latency: float = 1e-7, bw: float = 1e9, tdp: float = 100
    ) -> None:
        '''Set the parameters for the decoder hardware:

        Args
        ----
            name (str):
                name of decoder hardware
            speedup (float):
                scaling speedup factor how fast the hardware can run the
                decoder
            parallelism (int):
                number of decoders that can run in parallel on this hardware
            latency (float):
                latency (in s) of the network coupling qpu <> decoder
            bw (float):
                bandwidth in bit/s of the network coupling qpu <> decoder
            tdp (float):
                thermal desing power (in Watts) used by hardware
        '''
        self.name = name
        self.speedup = speedup
        self.parallelism = parallelism
        self.latency = latency
        self.bw = bw
        self.tdp = tdp


_cpu = _DecoderHardware(
    'cpu',
    speedup=1,             # 1x: native execution speed
    parallelism=128,       # number of threads
    latency=125e-9,        # based on PCIe 4.0 (in s)
    bw=64e9,               # based on PCIe 4.0 (in bps)
    tdp=280                # based on PM AMD Milan (in W)
)


@dataclass
class _Hardware:
    cpu: _DecoderHardware = _cpu


hardware = _Hardware()


@dataclass
class _Decoder:
    software: _DecoderAlgorithm
    hardware: _DecoderHardware


@dataclass
class _Decoders:
    cpu_linear: _Decoder = _Decoder(software=_linear, hardware=_cpu)
    cpu_quadratic: _Decoder = _Decoder(software=_quadratic, hardware=_cpu)


decoders = _Decoders()
