#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''Models of quantum error correction.'''
from typing import Callable
import numpy as np


class SurfaceCode:
    '''Surface code properties.'''
    def __init__(
            self, name: str, compute_number_of_tiles: Callable,
            cycles_per_magic_state: int, number_of_distillation_tiles: int
    ) -> None:
        self.name = name
        self.compute_number_of_tiles = compute_number_of_tiles
        self.cycles_per_magic_state = cycles_per_magic_state
        self.number_of_distillation_tiles = number_of_distillation_tiles

    def compute_total_number_of_tiles(self, n: int):
        '''Compute the total number of tiles (incl. data qubits, ancilla qubits
        and magic state distillation qubits) to support a computation on n
        logical qubits.

        Args
        ----
            n (Int):
                number of logical qubits
        '''
        return self.compute_number_of_tiles(n) + \
            self.number_of_distillation_tiles


compact_code = SurfaceCode(
    name='compact',
    compute_number_of_tiles=lambda n: 1.5*n + 3,
    cycles_per_magic_state=11,
    number_of_distillation_tiles=11
)


intermediate_code = SurfaceCode(
    name='intermediate',
    compute_number_of_tiles=lambda n: 2*n + 4,
    cycles_per_magic_state=5,
    number_of_distillation_tiles=22
)


fast_code = SurfaceCode(
    name='fast',
    compute_number_of_tiles=lambda n: 2*n + np.sqrt(8*n) + 1,
    cycles_per_magic_state=1,
    number_of_distillation_tiles=132
)
