#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''Useful code snippets.'''
import itertools
import numpy as np
import matplotlib as mpl


def decoding_error_rate_heuristic(ep: float, d: int):
    '''Returns the common heuristic for logical decoding error rate for a
    surface code of code distance `d` running  on a device with physical
    error rates `ep`.

    Args
    ----
        ep (Float):           physical error rate [Hz]
        d (Int):              code distance

    Returns
    -------
        el (Float): logical error rate [Hz]

    References:
        * Austin G. Fowler and Craig Gidney. 2019.
        Low overhead quantum computation using lattice
        surgery.

    '''
    return 0.1 * (100 * ep)**((d+1)/2.0)


def data_rate_heuristic(ep: float, d: int, tcyc: float):
    '''Returns the data rate heuristic for the detector graph data generation
    as identified in the paper.

    Args
    ---
        ep (Float):         physical error rate [Hz]
        d (Int):            code distance
        tcyc (Float):       code cycle time [s]

    Returns
    -------
        Rl (Float):         detector graph data generation rate (in bps)
    '''
    return 4 * d**2 * ep / tcyc


def round_up_to_odd(x: float):
    '''Round up to the nearest odd integer.

    Args
    ----
        x (Float): value to be rounded
    '''
    return np.ceil(x) // 2 * 2 + 1


def enumerated_product(*args):
    yield from zip(
        itertools.product(*(range(len(x)) for x in args)),
        itertools.product(*args)
    )


def load_plot_configuration():
    '''Function to configure the matplotlib rcParams and get my default paper
    color scheme.

    Returns
    -------
        colors (Iterable):
            tuple of 7 rgb tuples for color scheme
    '''
    mpl.rcParams['font.serif'] = [
        'cmr10', 'Computer Modern Serif', 'DejaVu Serif'
    ]
    mpl.rcParams['font.family'] = 'serif'
    mpl.rcParams['font.size'] = 16
    mpl.rcParams['axes.formatter.use_mathtext'] = True
    mpl.rcParams['text.usetex'] = True
    mpl.rcParams['text.latex.preamble'] = '\\usepackage{amsmath} \\usepackage{amssymb}'  # noqa
    colors = (
        (0.0000, 0.4470, 0.7410),  # myblue
        (0.8500, 0.3250, 0.0980),  # myred
        (0.9290, 0.6940, 0.1250),  # myorange
        (0.4940, 0.1840, 0.5560),  # mypurple
        (0.4660, 0.6740, 0.1880),  # mygreen
        (0.3010, 0.7450, 0.9330),  # mylightblue
        (0.6350, 0.0780, 0.1840),  # mydarkred
    )
    return colors
