import sys
import h5py
from pathlib import Path
import numpy as np
import numpy.typing as npt
import matplotlib.pyplot as plt
from copy import deepcopy
sys.path.insert(0, '../')
from qresourceanalyzer.analyzer import (  # noqa: E402
    QuantumSystem,
    compute_data_rate,
    logical_clock_speed,
    t_cycle_time,
    hscx_cycle_time,
    optimize_interruption_time_and_code_distance,
)
from qresourceanalyzer.util import enumerated_product, load_plot_configuration  # noqa
import qresourceanalyzer as qra  # noqa: E402


def main(load_file: str = None):
    # ----------------------------------------------------------------------- #
    # Script settings.
    data_path = "data/"
    distances = np.arange(5, 22, 2)
    q = 1000
    interruption_time = 100
    # ----------------------------------------------------------------------- #
    circuit = qra.qproblems.LogicalCircuit("tmp", qubits=0, tgates=0)
    qpu = qra.qdevices.superconducting
    code = qra.qec.fast_code
    dec_algo = qra.qdecoding.algorithms.sparseblossom
    dec_hw = qra.qdecoding.hardware.cpu
    qsystem = QuantumSystem(
        circuit=circuit,
        qpu=qpu,
        code=code,
        dec_algo=dec_algo,
        dec_hw=dec_hw
    )
    if load_file is None:
        load_file = f'{data_path}q_{q}_results.h5'
    if not Path(load_file).is_file():
        # data set does not exist, run the experiment
        print('No data found! Running benchmark experiment...')
        generate_and_store_data(
            load_file, interruption_time, q, qsystem, distances)
    generate_figure(load_file)


def generate_and_store_data(
        fname: str, M: float, q: int,
        qsystem: QuantumSystem, distances: npt.ArrayLike
):
    print('Computing data...')
    # HSTCX breakdown data
    t_t = np.zeros_like(distances, dtype=np.float64)
    t_hscx = np.zeros_like(distances, dtype=np.float64)
    t_msd = np.zeros_like(distances, dtype=np.float64)
    t_route = np.zeros_like(distances, dtype=np.float64)
    t_dec = np.zeros_like(distances, dtype=np.float64)
    t_dec_avg = np.zeros_like(distances, dtype=np.float64)
    t_latency = np.zeros_like(distances, dtype=np.float64)
    hz = np.zeros_like(distances, dtype=np.float64)
    hz_avg = np.zeros_like(distances, dtype=np.float64)
    raw_rate = np.zeros_like(distances, dtype=np.float64)
    compr_rate = np.zeros_like(distances, dtype=np.float64)

    _circuit = qra.qproblems.LogicalCircuit(
        name="tmp", qubits=q, tgates=1
    )
    _qsystem = deepcopy(qsystem)
    _qsystem.circuit = _circuit

    for i, d in enumerate(distances):
        # Compute interruption time
        t_t[i], t_msd[i], t_route[i], t_dec[i], t_latency[i] = \
            t_cycle_time(d=d, M=M, qsystem=_qsystem)
        t_dec_avg[i] = 2 * 6.1e-9 * d**4.3  # the average decoder performance
        t_hscx[i] = hscx_cycle_time(d, _qsystem)
        hz[i] = logical_clock_speed(d=d, M=M, qsystem=_qsystem)
        hzz = t_msd[i] + t_route[i] + t_latency[i] + t_hscx[i] + t_dec_avg[i]
        hz_avg[i] = 1.0/hzz
        raw_rate[i], compr_rate[i] = compute_data_rate(d, _qsystem)
    # Save to file
    print("saving data to hdf5...")
    hf = h5py.File(fname, 'w')
    # store data
    hf.create_dataset('q', data=q)
    hf.create_dataset('distances', data=distances)
    hf.create_dataset('t_t', data=t_t)
    hf.create_dataset('t_hscx', data=t_hscx)
    hf.create_dataset('t_msd', data=t_msd)
    hf.create_dataset('t_route', data=t_route)
    hf.create_dataset('t_dec', data=t_dec)
    hf.create_dataset('t_dec_avg', data=t_dec_avg)
    hf.create_dataset('t_latency', data=t_latency)
    hf.create_dataset('hz', data=hz)
    hf.create_dataset('hz_avg', data=hz_avg)
    hf.create_dataset('raw_rate', data=raw_rate)
    hf.create_dataset('compr_rate', data=compr_rate)
    hf.close()


def generate_figure(fname: str):
    # Load data from file
    print('Load data...')
    hf = h5py.File(fname, 'r')
    q = np.array(hf.get('q'))  # noqa
    distances = np.array(hf.get('distances'))
    t_t = np.array(hf.get('t_t'))  # noqa
    t_hscx = np.array(hf.get('t_hscx'))
    t_msd = np.array(hf.get('t_msd'))
    t_route = np.array(hf.get('t_route'))
    t_dec = np.array(hf.get('t_dec'))
    t_dec_avg = np.array(hf.get('t_dec_avg'))
    t_latency = np.array(hf.get('t_latency'))
    hz = np.array(hf.get('hz'))
    hz_avg = np.array(hf.get('hz_avg'))
    raw_rate = np.array(hf.get('raw_rate'))
    compr_rate = np.array(hf.get('compr_rate'))

    my_colors = load_plot_configuration()
    fig = plt.figure(constrained_layout=True, figsize=(15, 4))
    axs = fig.subplot_mosaic(
        [['Left', 'Mid', 'Right']],
        gridspec_kw={'width_ratios': [1, 1, 1]})

    # -- Left Plot: Stack Plot -- #
    ax = axs['Left']

    ax.stackplot(
        distances, t_hscx, t_msd, t_route, t_dec_avg,  t_latency,
        labels=['HSCX', 'MSD', 'Route', 'Decode', 'Latency'],
        colors=my_colors
    )

    ax.legend(loc="upper left")
    ax.set_xlabel(r"Code Distance ($d$)")
    ax.set_xticks(distances[::2])
    ax.set_xticklabels(distances[::2].astype(int), rotation=45, ha='right')
    ax.set_ylabel(r"t [s]")
    ax.set_xlim([distances[0], distances[-1]])
    ax.set_title('Logical operation time')
    ax.set_yscale("log")

    # -- Middle Plot: Logical Clock Speed -- #
    ax = axs['Mid']

    ax.plot(distances, hz_avg, color=my_colors[0], linewidth=3)
    ax.grid()
    ax.set_xlabel(r"Code Distance ($d$)")
    ax.set_xticks(distances[::2])
    ax.set_xticklabels(distances[::2].astype(int), rotation=45, ha='right')
    ax.set_ylabel(r"Clock speed [Hz]")
    ax.set_xlim([distances[0], distances[-1]])
    ax.set_ylim([100, 10000])
    ax.set_title('Logical clock speed')
    ax.set_yscale("log")
    # --- Right Plot: Data Generation Rate --- #
    ax = axs['Right']

    ax.plot(
        distances, raw_rate/1e6, color=my_colors[0], linewidth=3,
        label="Raw Syndrome"
    )
    ax.plot(
        distances, compr_rate/1e6, color=my_colors[1], linewidth=3,
        label="Compressed Syndrome"
    )
    ax.fill_between(distances, raw_rate/1e6, compr_rate/1e6,
                    color='k', alpha=0.2)
    ax.grid()
    ax.legend(loc="upper left")
    ax.set_yscale("log")
    ax.set_xlabel(r"Code Distance ($d$)")
    ax.set_xticks(distances[::2])
    ax.set_xticklabels(distances[::2].astype(int), rotation=45, ha='right')
    ax.set_ylabel(r"Data Rate [Mbps]")
    ax.set_xlim([distances[0], distances[-1]])
    ax.set_title('Data generation rate')
    plt.show()


if __name__ == "__main__":
    main(load_file=None)
