import sys
import h5py
import stim
import pymatching
from time import time
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np
import numpy.typing as npt
from pathlib import Path
sys.path.insert(0, '../')
from qresourceanalyzer.util import load_plot_configuration # noqa
from qresourceanalyzer.qdecoding import algorithms # noqa


def main(load_file: str = None):
    '''Main script function'''
    # -------------------------------------------------------------------------
    distances = np.arange(3, 53, 4)
    phys_error = 1e-3
    num_windows = 2500  # 500, 2500
    average = False     # True, False
    data_path = "data/"
    # -------------------------------------------------------------------------
    # https://www.hpcwire.com/2021/03/15/amd-launches-epyc-milan-with-19-skus-for-hpc-enterprise-and-hyperscale/  # noqa
    proc_spec_epyc = {
        'part_of_peak': 0.1,  # part of peak performance used by decoding
        'ncores': 64,
        'peak_flops': 3.58e12
    }
    if load_file is None:
        load_file = generate_file_name(
            data_path, distances, phys_error, num_windows, average
        )
    if not Path(load_file).is_file():
        # data set does not exist, run the experiment
        print('No data found! Running benchmark experiment...')
        generate_and_store_data(
            load_file, distances, phys_error, num_windows, average
        )
        print('Experiment completed...')
    print('Generating figure...')
    if average:
        generate_average_plot(load_file, proc_spec_epyc)
    else:
        generate_distribution_plot(load_file, proc_spec_epyc)


def generate_file_name(
        path: str, distances: npt.ArrayLike, phys_error: float,
        num_windows: int, average: bool = True
):
    '''Generate the file name'''
    if average:
        return (
            f'{path}time_average_decode_v_codedist_d_{distances[0]}_'
            f'{distances[-1]}_ep_{phys_error}_windows_{num_windows}.h5'
        )
    return (
        f'{path}time_distribution_decode_v_codedist_d_{distances[0]}_'
        f'{distances[-1]}_ep_{phys_error}_windows_{num_windows}.h5'
    )


def time_surface_code_circuit(
        distance: int, p: float, num_windows: int = 10000, average: bool = True
) -> float:
    '''Time the PyMatching benchmark per window.

    Args
    ----
        distance (int):
            code distance
        p (float):
            physical error rate
        num_windows (int):
            total number of windows of length 3*d to decode
        average (bool):
            report the average time over all windows or report the time per
            window
    https://github.com/oscarhiggott/PyMatching/tree/master/benchmarks/surface_codes  # noqa
    '''

    circuit = stim.Circuit.generated(
        "surface_code:rotated_memory_x",
        distance=distance,
        rounds=distance,
        after_clifford_depolarization=p,
        before_round_data_depolarization=p,
        before_measure_flip_probability=p,
        after_reset_flip_probability=p
    )
    window_size = 3 * distance
    num_shots = num_windows * window_size
    print(f'code distance: {distance}, shots: {num_shots}')
    dem = circuit.detector_error_model(decompose_errors=True)
    matching = pymatching.Matching.from_detector_error_model(dem)
    sampler = circuit.compile_detector_sampler()
    shots, actual_observables = sampler.sample(
        shots=num_shots, separate_observables=True
    )
    # Decode one shot first
    matching.decode_batch(shots[0:1, :])
    if average:
        # Now time decoding the batch
        t0 = time()
        matching.decode_batch(shots)
        t1 = time()
        seconds_per_window = (t1-t0)/num_windows
        return seconds_per_window
    seconds_per_window = np.zeros(num_windows)
    for i in range(num_windows):
        # Decode each window
        t0 = time()
        matching.decode_batch(shots[i*window_size:(i+1)*window_size, :])
        t1 = time()
        seconds_per_window[i] = (t1 - t0)
    return seconds_per_window


def generate_and_store_data(
        fname: str,  distances: npt.ArrayLike, phys_error: float,
        num_windows: int, average: bool = True
):
    # run the PyMatching benchmark
    print("running the benchmark...")
    timings = np.array(
        [time_surface_code_circuit(d, phys_error, num_windows, average) for d
         in distances]
    )
    # save data
    print("saving data to hdf5...")
    hf = h5py.File(fname, 'w')
    # store meta data
    hf.create_dataset('distances', data=distances)
    hf.create_dataset('phys_error', data=phys_error)
    hf.create_dataset('num_windows', data=num_windows)
    hf.create_dataset('average', data=average)
    # store results
    hf.create_dataset('timings', data=timings)
    hf.close()


def compute_exp_fit(x, y):
    fit = np.polyfit(np.log(x), np.log(y), 1)
    alpha, beta = np.exp(fit[1]), fit[0]
    y_fit = alpha*np.power(x, beta)
    float_str = "{0:.2g}".format(alpha)
    if "e" in float_str:
        base, exponent = float_str.split("e")
        label = r"${0} \cdot 10^{{{1}}} \times d^{{{2:.1f}}}$".format(
            base, int(exponent), beta
        )
    else:
        label = r"${0} \times d^{{{1:.1f}}}$".format(float_str, beta)
    return alpha, beta, y_fit, label


def generate_average_plot(fname: str, proc_spec: dict):
    # Load the data set
    hf = h5py.File(fname, 'r')
    distances = np.array(hf.get('distances'))
    timings = np.array(hf.get('timings'))
    phys_error = np.array(hf.get('phys_error'))
    num_windows = np.array(hf.get('num_windows'))
    average = bool(hf.get('average'))
    if not average:
        raise ValueError("Wrong data set! Expected average data.")
    print('Generating figure...')
    print(f'num_windows: {num_windows}')
    print(f'physical error: {phys_error}')
    my_colors = load_plot_configuration()
    # ----------------------------------------------------------------------- #

    fig = plt.figure(figsize=(11, 5))
    axs = fig.subplot_mosaic(
        [['Left', 'Middle']],
        gridspec_kw={'width_ratios': [1, 1]})

    # --- Left Plot: Timings -- #
    ax = axs['Left']
    # compute fit
    alpha, beta, timings_fit, my_label = compute_exp_fit(distances, timings)
    print(f'best fit: alpha: {alpha}, beta: {beta}')
    ax.plot(distances, timings, color=my_colors[0], marker="o", linewidth=0,
            markersize=8, mew=3)
    ax.plot(
        distances, timings_fit,
        color='k', linestyle='dashed', label=my_label
    )
    ax.legend(loc="upper left")
    ax.grid()
    ax.set_xlabel(r"Code Distance ($d$)")
    ax.set_ylabel(r"Decoding Time ($s$)")
    ax.set_yscale("log")
    ax.set_xscale("log")
    ax.set_xlim([distances[0], distances[-1]])
    xtick_dist = distances[[0, 1, 3, 7, 12]]
    ax.set_xticks(xtick_dist)
    ax.set_xticklabels(xtick_dist.astype(int), rotation=45, ha='right')

    # --- Middle Plot: tot OP -- #
    ax = axs['Middle']
    # compute data
    part_of_peak = proc_spec['part_of_peak']
    ncores = proc_spec['ncores']
    peak_flops = proc_spec['peak_flops']
    tot_ops = part_of_peak * timings * (peak_flops / ncores)
    alpha, beta, tot_ops_fit, my_label = compute_exp_fit(distances, tot_ops)
    ax.plot(
        distances, tot_ops, color=my_colors[0], marker="o", linewidth=0,
        markersize=8, mew=3
    )
    ax.plot(
        distances, tot_ops_fit,
        color='k', linestyle='dashed', label=my_label
    )
    ax.legend(loc="upper left")
    ax.grid()
    ax.set_xlabel(r"Code Distance ($d$)")
    ax.set_ylabel(r"Estimated Equivalent FLOP")
    ax.set_yscale("log")
    ax.set_xscale("log")
    ax.set_xlim([distances[0], distances[-1]])
    ax.set_xticks(xtick_dist)
    ax.set_xticklabels(xtick_dist.astype(int), rotation=45, ha='right')
    plt.tight_layout()
    plt.show()

def generate_distribution_plot(fname: str, peak_flops: float):
    # Load the data set
    hf = h5py.File(fname, 'r')
    distances = np.array(hf.get('distances'))
    timings = np.array(hf.get('timings'))
    phys_error = np.array(hf.get('phys_error'))
    num_windows = np.array(hf.get('num_windows'))
    average = bool(hf.get('average'))
    if not average:
        raise ValueError("Wrong data set! Expected distribution data.")
    print('Generating figure...')
    print(f'num_windows: {num_windows}')
    print(f'physical error: {phys_error}')

    # Generate synthetic data based on the decoder model.
    synth_data = []
    for d in distances:
        synth_data.append(
            algorithms.sparseblossom.generate_runtimes(
                num_windows, phys_error, d
            )
        )
    my_colors = load_plot_configuration()

    fig, ax = plt.subplots()
    # Experimental data
    violin_parts = ax.violinplot(
        timings.tolist()[1:], positions=distances[1:], widths=3,
        showmeans=True, showextrema=False, showmedians=False
    )
    for vp in violin_parts['bodies']:
        vp.set_facecolor(my_colors[0])
        vp.set_facecolor(my_colors[0])
        vp.set_linewidth(1)
        vp.set_alpha(0.5)
    # Synthetic data
    violin_parts = ax.violinplot(
        synth_data[1:], positions=distances[1:]+1, widths=3, showmeans=True,
        showextrema=False, showmedians=False
    )
    for vp in violin_parts['bodies']:
        vp.set_facecolor(my_colors[1])
        vp.set_facecolor(my_colors[1])
        vp.set_linewidth(1)
        vp.set_alpha(0.5)
    # Legend
    patches = (
        mpatches.Patch(color=my_colors[0]),
        mpatches.Patch(color=my_colors[1])
    )
    labels = ["PyMatching Timing", "Model Eqs. (12), (13)"]

    ax.legend(
       patches, labels, loc="lower right"
    )
    ax.set_yscale("log")
    ax.set_xlabel(r"Code Distance ($d$)")
    ax.set_ylabel(r"Decoding Time ($s$)")
    ax.xaxis.set_ticks(distances[1:])
    ax.set_xticklabels(distances[1:])
    ax.grid()
    plt.show()


if __name__ == "__main__":
    main(load_file=None)
