import sys
import h5py
import stim
import numpy as np
import numpy.typing as npt
import matplotlib.pyplot as plt
from pathlib import Path
sys.path.insert(0, '../')
from qresourceanalyzer.util import load_plot_configuration, enumerated_product  # noqa


def main(load_file: str = None):
    '''Main script function:
        * if load_file is specified: it will load the data from the hdf5 file
        and generate the plot (include relative path)
        * if load_file is None: data will be generated, stored and plotted
    '''
    code = "surface_code:rotated_memory_x"
    distances = np.arange(3, 53, 4)
    phys_errors = np.logspace(-2, -4, 3)
    shots = 10_000
    sample_freq = 1e6
    data_path = "data/"
    if load_file is None:
        load_file = generate_file_name(
            data_path, code, distances, phys_errors, shots
        )
    if not Path(load_file).is_file():
        # data set does not exist, run the experiment
        print('No data found! Running benchmark experiment...')
        generate_and_store_data(load_file, code, distances, phys_errors, shots)
    print('Generating figure...')
    generate_plot(load_file, sample_freq)


def generate_file_name(
        path: str, code: str, distances: npt.ArrayLike,
        phys_errors: npt.ArrayLike, shots: int
):
    '''Generate the filename.'''
    return (
        f'{path}data_codedist_v_datarate_{code}_d_{distances[0]}_'
        f'{distances[-1]}_ep_{phys_errors[0]}_{phys_errors[-1]}_shots_{shots}'
        '.h5'
    )


def generate_and_store_data(
        fname: str, code: str, distances: npt.ArrayLike,
        phys_errors: npt.ArrayLike, shots: int = 1_000
):
    '''
    Generate and save the figure data.
    '''
    syndrome_data = np.zeros((len(distances), len(phys_errors)))
    compr_data_means = np.zeros((len(distances), len(phys_errors)))
    compr_data_std = np.zeros((len(distances), len(phys_errors)))

    for (i, j), (d, ep) in enumerated_product(distances, phys_errors):
        print('********************************')
        print(f'* code distance: {d}')
        print(f'* error rate: {ep}')
        print('********************************')
        circuit = stim.Circuit.generated(
            code,
            rounds=d,
            distance=d,
            before_round_data_depolarization=ep,
            before_measure_flip_probability=ep
        )

        sampler = circuit.compile_sampler()
        samples = sampler.sample(shots=shots)
        syndrome_data[i, j] = len(samples[0])
        data = np.count_nonzero(samples, axis=1)

        sampler = circuit.compile_detector_sampler()
        samples = sampler.sample(shots=shots)
        data = np.count_nonzero(samples, axis=1)
        compr_data_means[i, j] = np.mean(data)
        compr_data_std[i, j] = np.std(data)
    # save data
    print("saving data to hdf5...")
    hf = h5py.File(fname, 'w')
    # store meta data
    hf.create_dataset('distances', data=distances)
    hf.create_dataset('phys_errors', data=phys_errors)
    hf.create_dataset('shots', data=shots)
    # store experimental results
    hf.create_dataset('syndrome_data', data=syndrome_data)
    hf.create_dataset('compr_data_means', data=compr_data_means)
    hf.create_dataset('compr_data_std', data=compr_data_std)
    hf.close()


def generate_plot(fname: str, freq: float):
    # Load the data set
    hf = h5py.File(fname, 'r')
    distances = np.array(hf.get('distances'))
    phys_errors = np.array(hf.get('phys_errors'))
    syndrome_data = np.array(hf.get('syndrome_data'))
    compr_data_means = np.array(hf.get('compr_data_means'))
    compr_data_std = np.array(hf.get('compr_data_std'))

    # Load Plot settings.
    my_colors = load_plot_configuration()
    markers = ["+", "^", "x"]

    def my_syndrome_approx(d):
        return d**2 * freq

    def my_compr_approx(d, ep):
        return 4 * d**2 * ep * freq

    fig, ax = plt.subplots()
    # syndromes
    ax.plot(
        distances, freq*syndrome_data[:, 0] / distances / 1e6,
        color=my_colors[0], marker="o", linestyle="dashed", markersize=8,
        mew=3, linewidth=3
    )
    ax.plot(
            distances, [my_syndrome_approx(d) / 1e6 for d in distances],
            color='k', linewidth=1.5,
            label=r'$d^2 \cdot t_{\text{cycle}}^{-1}$'
        )

    for j in range(3):
        ax.plot(
            distances, freq*compr_data_means[:, j]/distances / 1e6,
            color=my_colors[j+1],  marker=markers[j], markersize=8,  mew=3,
            linewidth=3, linestyle="dashed"
        )
        ax.fill_between(
            distances,
            freq*(compr_data_means[:, j] + compr_data_std[:, j])/distances/1e6,
            freq*(compr_data_means[:, j] - compr_data_std[:, j])/distances/1e6,
            color=my_colors[j+1], alpha=0.25
        )
        ax.plot(
            distances,
            [my_compr_approx(d, phys_errors[j]) / 1e6 for d in distances],
            linestyle='dotted', color='k', linewidth=1.5,
            label=(r'$ 4 \cdot e_{\text{p}} \cdot d^2 \cdot t_{\text{cycle}}^{-1}$' if j == 0 else None)  # noqa
        )
    ax.set_yscale("log")
    ax.set_xlabel(r"Code Distance ($d$)")
    ax.set_ylabel('Data Rate [Mbps]')
    ax.set_xlim([distances[0], distances[-1]])
    ax.set_ylim([1e-4, 1e4])
    ax.grid()
    ax.set_xticks(distances[::1])
    ax.set_xticklabels(distances[::1], rotation=45, ha='right')
    ax.legend(loc="lower right")
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    main(load_file=None)
