import sys
import h5py
from pathlib import Path
import numpy as np
import numpy.typing as npt
from typing import Iterable
import matplotlib.pyplot as plt
from matplotlib import cm
from copy import deepcopy
sys.path.insert(0, '../')
from qresourceanalyzer.analyzer import (  # noqa: E402
    QuantumSystem,
    compute_nwindows,
    compute_runtime,
    optimize_interruption_time_and_code_distance,
)
from qresourceanalyzer.util import enumerated_product, load_plot_configuration  # noqa
import qresourceanalyzer as qra  # noqa: E402


def main(load_file: str = None):
    # ----------------------------------------------------------------------- #
    # Script settings.
    data_path = "data/"
    marked_problems = [
        qra.qproblems.hub8, qra.qproblems.hub20, qra.qproblems.p450
    ]
    #
    error_budget = 0.5
    distances = np.arange(3, 29, 2)
    npoints = 100             # 1D grid points
    min_q, max_q = 10, 2500   # linear scale
    min_t, max_t = 5, 10      # log scale
    nqubits = np.linspace(min_q, max_q, npoints, dtype=int).astype(float)
    ntgates = np.logspace(min_t, max_t, npoints, dtype=int).astype(float)
    # ----------------------------------------------------------------------- #
    circuit = qra.qproblems.LogicalCircuit("tmp", qubits=0, tgates=0)
    qpu = qra.qdevices.superconducting
    code = qra.qec.fast_code
    dec_algo = qra.qdecoding.algorithms.sparseblossom
    dec_hw = qra.qdecoding.hardware.cpu
    qsystem = QuantumSystem(
        circuit=circuit,
        qpu=qpu,
        code=code,
        dec_algo=dec_algo,
        dec_hw=dec_hw
    )
    if load_file is None:
        load_file = generate_file_name(
            data_path, (min_q, max_q), (min_t, max_t), npoints
        )
    if not Path(load_file).is_file():
        # data set does not exist, run the experiment
        print('No data found! Running benchmark experiment...')
        generate_and_store_data(
            load_file, error_budget, nqubits, ntgates, qsystem, distances
        )
        print('Experiment completed...')
    generate_contour_plot(load_file, marked_problems)


def generate_file_name(
        path: str, q_range, t_range, npoints
):
    '''Generate the file name'''
    return (
        f'{path}contour_data_q_{q_range[0]}_{q_range[-1]}_t_{t_range[0]}_'
        f'{t_range[-1]}_n_{npoints}.h5'
    )


def generate_and_store_data(
        fname: str, error_budget, nqubits: npt.ArrayLike,
        ntgates: npt.ArrayLike, qsystem: QuantumSystem, d_range: npt.ArrayLike
):
    # Compute all data
    print('Computing plot data....')
    NQ, NT = np.meshgrid(nqubits, ntgates)

    code_distances = np.zeros_like(NQ)
    interruption_times = np.zeros_like(NQ)
    run_times = np.zeros_like(NQ)
    estimated_flops = np.zeros_like(NQ)
    estimated_ops = np.zeros_like(NQ)
    num_windows = np.zeros_like(NQ)

    for (i, j), (q, t) in enumerated_product(nqubits, ntgates):
        _circuit = qra.qproblems.LogicalCircuit(
            name="tmp", qubits=q, tgates=t
        )
        _qsystem = deepcopy(qsystem)
        _qsystem.circuit = _circuit
        interruption_times[j, i], code_distances[j, i], _, _ = \
            optimize_interruption_time_and_code_distance(
                error_budget=error_budget, qsystem=_qsystem, d_range=d_range
            )
        if code_distances[j, i]:
            run_times[j, i] = np.log10(compute_runtime(
                code_distances[j, i], M=interruption_times[j, i],
                qsystem=_qsystem
            ))
        else:
            run_times[j, i] = np.nan  # should not happen
        num_windows[j, i] = compute_nwindows(
            d=code_distances[j, i], M=interruption_times[j, i], qsystem=qsystem
        )
        # fitted operation model
        estimated_ops[j, i] = (
            2 * num_windows[j, i] * 34 * (code_distances[j, i]**4.3)
        )  # for 1 HSTCX-cycle
        part_of_peak = 0.1
        estimated_flops[j, i] = (
            estimated_ops[j, i] / (interruption_times[j, i]*1e-6*part_of_peak)
        )
    # Save to file
    print("saving data to hdf5...")
    hf = h5py.File(fname, 'w')
    # store meta data
    hf.create_dataset('nqubits', data=nqubits)
    hf.create_dataset('ntgates', data=ntgates)
    # store results
    hf.create_dataset('NQ', data=NQ)
    hf.create_dataset('NT', data=NT)
    hf.create_dataset('code_distances', data=code_distances)
    hf.create_dataset('interruption_times', data=interruption_times)
    hf.create_dataset('run_times', data=run_times)
    hf.create_dataset('num_windows', data=num_windows)
    hf.create_dataset('estimated_ops', data=estimated_ops)
    hf.create_dataset('estimated_flops', data=estimated_flops)
    hf.close()


def generate_contour_plot(fname: str, marked_problems: Iterable):
    # Load data from file
    print('Load data...')
    hf = h5py.File(fname, 'r')
    NQ = np.array(hf.get('NQ'))
    NT = np.array(hf.get('NT'))
    code_distances = np.array(hf.get('code_distances'))
    interruption_times = np.array(hf.get('interruption_times'))
    run_times = np.array(hf.get('run_times'))
    num_windows = np.array(hf.get('num_windows'))
    estimated_flops = np.array(hf.get('estimated_flops'))
    estimated_ops = np.array(hf.get('estimated_ops'))  # noqa

    print('Generating plot...')
    my_colors = load_plot_configuration()
    color_map = cm.bone
    fig = plt.figure(figsize=(21, 8))
    layout = [
        ['M', 'D', 'C', 'C'],
        ['W', 'T', 'C', 'C'],
    ]
    axs = fig.subplot_mosaic(layout)

    # ----------------------------------------------------------------------- #
    # Subfig D: Code distance
    # ----------------------------------------------------------------------- #
    ax = axs['D']
    level_lines = [5, 7, 9, 11, 13, 15, 17, 19, 21]
    CS = ax.contourf(
        NQ, NT, code_distances,
        levels=level_lines, cmap=color_map
    )
    cbar = fig.colorbar(CS, ax=ax)
    cbar.set_ticks(level_lines)
    cbar.set_ticklabels(level_lines, rotation=45, ha='left')
    cbar.ax.set_ylabel(r'Code Distance ($d$)')

    # ----------------------------------------------------------------------- #
    # Subfig M: Interruption times
    # ----------------------------------------------------------------------- #
    ax = axs['M']
    interruption_times = np.log10(interruption_times * 1e-6)
    level_lines = [
        np.log10(3e-6), -5, np.log10(3e-5), -4, np.log10(3e-4),
        -3, np.log10(3e-3), -2
    ]
    level_labels = [
        3, 10, 30, 100, 300, 1000, 3000, 10000
        # r'1 $\mu s$', r'3 $\mu s$', r'10 $\mu s$', r'30 $\mu s$',
        # r'100 $\mu s$', r'300 $\mu s$', r'1000 $\mu s$', r'3000 $\mu s$'
    ]
    CS = ax.contourf(
        NQ, NT, interruption_times,
        levels=level_lines, cmap=color_map
    )
    cbar = fig.colorbar(CS, ax=ax)
    cbar.set_ticks(level_lines)
    cbar.set_ticklabels(level_labels)
    cbar.ax.set_ylabel(r'Interruption Time ($M$) [$\mu s$]')

    # ----------------------------------------------------------------------- #
    # Subfig W: windows
    # ----------------------------------------------------------------------- #
    ax = axs['W']
    CS = ax.contourf(
        NQ, NT, num_windows, cmap=color_map
    )
    cbar = fig.colorbar(CS, ax=ax)
    cbar.ax.set_ylabel(r'Number of Windows')
    # ----------------------------------------------------------------------- #
    # Subfig T: Run times
    # ----------------------------------------------------------------------- #
    ax = axs['T']

    time_vals = {
        'minute': 60,         # minute
        'hour': 3600,       # hour
        'day': 86400,      # day
        'week': 604800,     # week
        'month': 2.628e+6,   # month
        'year': 3.154e+7   #year
    }

    level_lines = [1, 2, 3, 4, 5, 6, 7, 8]
    level_labels = [
        r'$10^1$',
        r'$10^2$',
        r'$10^3$',
        r'$10^4$',
        r'$10^5$',
        r'$10^6$',
        r'$10^7$',
        r'$10^8$'
    ]
    CS = ax.contourf(
        NQ, NT, run_times, levels=level_lines, cmap=color_map
    )

    CS2 = ax.contour(
        NQ, NT, run_times,
        np.log10([v for v in time_vals.values()]),
        colors=my_colors, linewidths=(3,)
    )
    my_fmt = {k: v for k, v in zip(CS2.levels, time_vals.keys())}
    y_vals = [5e3, 3.5e5, 1e7, 1e8, 5e8, 3e9]
    my_manual = [
        (1200, y) for k, y in zip(np.linspace(0, 2000, 6), y_vals)
    ]

    ax.clabel(
        CS2, CS2.levels, manual=my_manual, inline=True, fmt=my_fmt,
        fontsize=16, colors=my_colors, inline_spacing=10
    )

    cbar = fig.colorbar(CS, ax=ax)
    cbar.set_ticks(level_lines)
    cbar.set_ticklabels(level_labels)
    cbar.ax.set_ylabel(r'Run Time [$s$]')

    # ----------------------------------------------------------------------- #
    # Subfig C: Compute
    # ----------------------------------------------------------------------- #
    ax = axs['C']
    level_lines = [13.66, 14, 14.33, 14.66, 15, 15.33]
    level_labels = [
        r'$4.6 \times 10^{13}$',
        r'$1.0 \times 10^{14}$',
        r'$2.1 \times 10^{14}$',
        r'$4.6 \times 10^{14}$',
        r'$1.0 \times 10^{15}$',
        r'$2.1 \times 10^{15}$',
    ]
    estimated_flops = np.log10(estimated_flops)
    CS = ax.contourf(
        NQ, NT, estimated_flops,
        levels=level_lines, cmap=cm.bone
    )
    CS2 = ax.contour(
        NQ, NT, estimated_flops,
        [14, 15],
        colors=[my_colors[1], my_colors[0]], linewidths=(3,)
    )
    ax.clabel(
        CS2, [14, 15],
        inline=True,
        fmt={14: '100 TFLOP/s', 15: '1 PFLOP/s'},
        fontsize=18,
        colors=[my_colors[1], my_colors[0]],
        inline_spacing=20
    )
    cbar = fig.colorbar(CS, ax=ax)
    cbar.set_ticks(level_lines)
    cbar.set_ticklabels(level_labels)
    cbar.ax.set_ylabel(r'Estimated Equivalent FLOP/s   $(\mathcal{C})$')

    # ----------------------------------------------------------------------- #
    # Add everything that plots have in common
    # ----------------------------------------------------------------------- #
    my_markers = ['+', '^', 'x']
    for ax in axs.values():
        ax.set_yscale("log")
        ax.set_xlabel(r"$Q_{\ell}$")
        ax.set_ylabel(r"$n_T$")
        for i, prob in enumerate(marked_problems):
            q = prob.qubits
            t = prob.tgates
            ax.plot(q, t, my_markers[i], mew=3, ms=8, color='r')
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    main(load_file=None)
